subroutine uv_shift_mosaic(line,error)
  use clean_def
  use clean_default
  use clean_arrays
  use phys_const
  use gkernel_types
  use gkernel_interfaces
  use gbl_message
  use imager_interfaces, except_this => uv_shift_mosaic
  !-------------------------------------------------------------------
  ! @ private
  !  MAPPING
  !   Support routine for command
  !       UV_SHIFT  [Offx Offy [Angle]]
  !
  !   Phase shift a Mosaic or Single Dish UV Table to a new
  !   common phase center and orientation
  !     Offx OffY are offsets in ArcSec 
  !               (default 0,0)
  !     Angle     is the final position angle from North in Degree
  !               (default: no change)
  !
  !   Also called implicitely (with no command line arguments, i.e.
  !   no change of Phase Center unless the UV Table is with
  !   Phase Offsets )
  !   by command UV_MAP for Mosaics
  !-------------------------------------------------------------------
  character(len=*), intent(in) :: line  ! Command line
  logical error                         ! Logical error flag
  !
  character(len=*), parameter :: rname='UV_SHIFT'
  character(len=80) :: mess
  type(projection_t) :: proj
  real(8) :: pos(2)
  real(4) :: cs(2), angnew, angold
  real(8), allocatable :: rpos(:,:)
  real(8) :: freq
  logical :: precise=.true.
  integer :: nc,nu,nv,i,ier,loff,moff
  !
  loff = huv%gil%column_pointer(code_uvt_loff)
  moff = huv%gil%column_pointer(code_uvt_moff)
  pos = 0
  !
  if (themap%nfields.eq.0) then
    call map_message(seve%i,rname,'UV data is a single field') 
  else
    if (loff.eq.0 .and. moff.eq.0) then
      call map_message(seve%w,rname,'Mosaic UV data is already in Pointing offsets')
      themap%nfields = +abs(themap%nfields)
      if (sic_narg(0).eq.0) return
    endif
    write(mess,'(A,I0,A)') 'Mosaic UV data has ',abs(themap%nfields),' fields'
    call map_message(seve%i,rname,mess)
    !
    ! Compute the mean offset as phase center position
    pos(1) = minval(themap%offxy(1,:)) + maxval(themap%offxy(1,:))
    pos(2) = minval(themap%offxy(2,:)) + maxval(themap%offxy(2,:))
    pos = pos/2*180*3600/pi  
  endif                  
  !
  angnew = huv%gil%pang*180/pi
  if (sic_narg(0).ne.0) then
    call sic_r8(line,0,1,pos(1),.false.,error)
    call sic_r8(line,0,2,pos(2),.true.,error)
    if (error) return
    call sic_r4(line,0,3,angnew,.false.,error)
  endif
  pos = pos*pi/180/3600
  angnew = angnew*pi/180
  !
  if (any(pos.ne.0d0)) then
    call gwcs_projec(huv%gil%a0,huv%gil%d0,-huv%gil%pang,huv%gil%ptyp,proj,error)
    call rel_to_abs (proj,pos(1),pos(2),huv%gil%a0,huv%gil%d0,1)
  endif
  !
  ! Compute observing frequency, and new phase center in wavelengths
  !
  if (precise) then
    nc = huv%gil%nchan
    allocate(rpos(2,nc),stat=ier)
    do i=1,huv%gil%nchan
      freq = gdf_uv_frequency(huv,dble(i))
      rpos(1:2,i) = - freq * f_to_k * pos(1:2)
    enddo
  else
    nc = 1
    allocate(rpos(2,1),stat=ier)
    freq = gdf_uv_frequency(huv)
    rpos(1:2,1) = - freq * f_to_k * pos(1:2)
  endif
  !
  ! Define the rotation
  cs  = [1.0,0.0]
  angold = huv%gil%pang
  huv%gil%pang = angnew
  cs(1)  =  cos(huv%gil%pang-angold)
  cs(2)  = -sin(huv%gil%pang-angold)
  !
  ! Recenter all channels, Loop over line table
  ! We work in place - We do not play with UV buffers
  !
  nu = huv%gil%dim(1)
  nv = huv%gil%nvisi
  call shift_uvdata (huv,nu,nv,duv,cs,nc,rpos)
  !
  ! Done if Single Field or already Pointing columns
  if (themap%nfields.ge.0) return
  !
  ! If not: Convert the LOFF - MOFF columns into Pointing columns...
  huv%gil%column_pointer(code_uvt_xoff) = loff
  huv%gil%column_pointer(code_uvt_yoff) = moff
  huv%gil%column_size(code_uvt_xoff) = 1
  huv%gil%column_size(code_uvt_yoff) = 1
  !
  huv%gil%column_pointer(code_uvt_loff) = 0
  huv%gil%column_pointer(code_uvt_moff) = 0
  huv%gil%column_size(code_uvt_loff) = 0
  huv%gil%column_size(code_uvt_moff) = 0
  !
  themap%nfields = abs(themap%nfields)
  !
end subroutine uv_shift_mosaic
!
